@@ -7,6 +7,8 @@
 
 #include <assert.h>
 
+#include "xkbcommon/xkbcommon.h"
+#include "atom.h"
 #include "x11-priv.h"
 
 /*
@@ -398,6 +400,9 @@ get_types(struct xkb_keymap *keymap, xcb_connection_t *conn,
             }
         }
 
+        /* Checked only when compiling a keymap from text */
+        type->required = true;
+
         xcb_xkb_key_type_next(&types_iter);
     }
 
@@ -421,6 +426,7 @@ get_sym_maps(struct xkb_keymap *keymap, xcb_connection_t *conn,
 
     keymap->min_key_code = reply->minKeyCode;
     keymap->max_key_code = reply->maxKeyCode;
+    keymap->num_keys = keymap->num_keys_low = reply->maxKeyCode + 1;
 
     ALLOC_OR_FAIL(keymap->keys, keymap->max_key_code + 1);
 
@@ -435,7 +441,7 @@ get_sym_maps(struct xkb_keymap *keymap, xcb_connection_t *conn,
         FAIL_UNLESS(key->num_groups <= ARRAY_SIZE(wire_sym_map->kt_index));
         ALLOC_OR_FAIL(key->groups, key->num_groups);
 
-        for (unsigned j = 0; j < key->num_groups; j++) {
+        for (xkb_layout_index_t j = 0; j < key->num_groups; j++) {
             FAIL_UNLESS(wire_sym_map->kt_index[j] < keymap->num_types);
             key->groups[j].type = &keymap->types[wire_sym_map->kt_index[j]];
 
@@ -459,6 +465,9 @@ get_sym_maps(struct xkb_keymap *keymap, xcb_connection_t *conn,
 
             FAIL_UNLESS((unsigned) syms_length == wire_sym_map->width * key->num_groups);
 
+            if (syms_length > 0)
+                key->explicit |= EXPLICIT_SYMBOLS;
+
             for (xkb_layout_index_t group = 0; group < key->num_groups; group++) {
                 for (xkb_level_index_t level = 0; level < wire_sym_map->width; level++) {
                     xcb_keysym_t wire_keysym = *syms_iter;
@@ -469,6 +478,9 @@ get_sym_maps(struct xkb_keymap *keymap, xcb_connection_t *conn,
                          * there may be an action set */
                         key->groups[group].levels[level].num_syms = 1;
                         key->groups[group].levels[level].s.sym = wire_keysym;
+                        /* Set capitalization transformation */
+                        key->groups[group].levels[level].upper =
+                            xkb_keysym_to_upper(wire_keysym);
                     }
 
                     syms_iter++;
@@ -586,17 +598,25 @@ get_explicits(struct xkb_keymap *keymap, xcb_connection_t *conn,
         key = &keymap->keys[wire->keycode];
 
         if ((wire->explicit & XCB_XKB_EXPLICIT_KEY_TYPE_1) &&
-            key->num_groups > 0)
+            key->num_groups > 0) {
             key->groups[0].explicit_type = true;
+            key->explicit |= EXPLICIT_TYPES;
+        }
         if ((wire->explicit & XCB_XKB_EXPLICIT_KEY_TYPE_2) &&
-            key->num_groups > 1)
+            key->num_groups > 1) {
             key->groups[1].explicit_type = true;
+            key->explicit |= EXPLICIT_TYPES;
+        }
         if ((wire->explicit & XCB_XKB_EXPLICIT_KEY_TYPE_3) &&
-            key->num_groups > 2)
+            key->num_groups > 2) {
             key->groups[2].explicit_type = true;
+            key->explicit |= EXPLICIT_TYPES;
+        }
         if ((wire->explicit & XCB_XKB_EXPLICIT_KEY_TYPE_4) &&
-            key->num_groups > 3)
+            key->num_groups > 3) {
             key->groups[3].explicit_type = true;
+            key->explicit |= EXPLICIT_TYPES;
+        }
         if (wire->explicit & XCB_XKB_EXPLICIT_INTERPRET) {
             key->explicit |= EXPLICIT_INTERP;
             /* Make all key groups have explicit actions too */
@@ -837,8 +857,13 @@ get_sym_interprets(struct xkb_keymap *keymap, xcb_connection_t *conn,
             sym_interpret->virtual_mod = NUM_REAL_MODS + wire->virtualMod;
 
         sym_interpret->repeat = (wire->flags & 0x01);
-        translate_action(&sym_interpret->action,
+        translate_action(&sym_interpret->a.action,
                          (xcb_xkb_action_t *) &wire->action);
+        sym_interpret->num_actions =
+            (sym_interpret->a.action.type != ACTION_TYPE_NONE);
+
+        /* Checked only when compiling a keymap from text */
+        sym_interpret->required = true;
 
         xcb_xkb_sym_interpret_next(&iter);
     }
@@ -893,9 +918,10 @@ get_type_names(struct xkb_keymap *keymap, struct x11_atom_interner *interner,
         uint8_t wire_num_levels = *n_levels_per_type_iter;
         struct xkb_key_type *type = &keymap->types[i];
 
-        /* Levels must have names. */
-        FAIL_UNLESS(type->num_levels == wire_num_levels);
+        /* Levels names are optional, but the maximum is the level count */
+        FAIL_UNLESS(type->num_levels >= wire_num_levels);
 
+        /* Allocate names for all levels, even if some names are missing */
         ALLOC_OR_FAIL(type->level_names, type->num_levels);
 
         x11_atom_interner_adopt_atom(interner, wire_type_name, &type->name);
@@ -903,6 +929,10 @@ get_type_names(struct xkb_keymap *keymap, struct x11_atom_interner *interner,
             x11_atom_interner_adopt_atom(interner, kt_level_names_iter[j],
                                          &type->level_names[j]);
         }
+        /* We may have received less names than there are levels: ensure
+         * missing names have a fallback. */
+        for (size_t j = wire_num_levels; j < type->num_levels; j++)
+            type->level_names[j] = XKB_ATOM_NONE;
 
         type->num_level_names = type->num_levels;
         kt_level_names_iter += wire_num_levels;
@@ -1157,7 +1187,7 @@ xkb_x11_keymap_new_from_device(struct xkb_context *ctx,
 
     if (device_id < 0 || device_id > 127) {
         log_err_func(ctx, XKB_LOG_MESSAGE_NO_ID,
-                     "illegal device ID: %d\n", device_id);
+                     "illegal device ID: %"PRId32"\n", device_id);
         return NULL;
     }
 
